
from inspect import iscoroutinefunction, Signature, Parameter
from types import MappingProxyType
from typing import Any, Callable, Optional

from cmdfns.invalid_argument import InvalidArgument


class Command:
    """
    Holds information about a command function.
    """

    def __init__(self, name: str, path: str, function: Callable[..., Any]):
        self.name = name
        self.path = path
        self.function = function

    def execute(self, positional_args: list[str],
                keyword_args: dict[str, str]) -> Any:
        parameters: MappingProxyType[str, Parameter] = \
            Signature.from_callable(self.function).parameters
        positional_args = \
            self._convert_positional_args(positional_args, parameters)
        keyword_args = self._convert_keyword_args(keyword_args, parameters)
        return self.function(*positional_args, **keyword_args)

    async def execute_async(self, positional_args: list[str],
                            keyword_args: dict[str, str]) -> Any:
        parameters: MappingProxyType[str, Parameter] = \
            Signature.from_callable(self.function).parameters
        positional_args = \
            self._convert_positional_args(positional_args, parameters)
        keyword_args = self._convert_keyword_args(keyword_args, parameters)
        if iscoroutinefunction(self.function):
            return await self.function(*positional_args, **keyword_args)
        else:
            return self.function(*positional_args, **keyword_args)

    def args_string(self) -> str:
        parameters: MappingProxyType[str, Parameter] = \
            Signature.from_callable(self.function).parameters
        parameter_strs: list[str] = []
        for name, parameter in parameters.items():
            if parameter.default is Parameter.empty:
                parameter_strs.append(name.upper())
            else:
                parameter_strs.append(f"[{name.upper()}]")
        return " ".join(parameter_strs)

    def _convert_positional_args(self, args: list[str],
                                 parameters: MappingProxyType[str, Parameter]
                                 ) -> list[Any]:
        converted_args: list[Any] = []
        parameters_list: list[Parameter] = list(parameters.values())
        for i in range(len(args)):
            arg = args[i]
            if i < len(parameters_list):
                parameter: Parameter = parameters_list[i]
                converted_arg: Any = None
                try:
                    converted_arg = self._convert(
                        arg, parameter.annotation)
                except Exception:
                    raise InvalidArgument(
                        f"Cannot convert '{arg}' to " +
                        f"{parameter.annotation.__name__} for parameter " +
                        f"'{parameter.name}' of command '{self.name}'")
                converted_args.append(converted_arg)
            else:
                converted_args.append(arg)
        return converted_args

    def _convert_keyword_args(self, args: dict[str, str],
                              parameters: MappingProxyType[str, Parameter]
                              ) -> dict[str, Any]:
        converted_args: dict[str, Any] = {}
        for name, arg in args.items():
            parameter: Optional[Parameter] = parameters.get(name)
            if parameter is not None:
                converted_arg: Any = None
                try:
                    converted_arg = self._convert(
                        arg, parameter.annotation)
                except Exception:
                    raise InvalidArgument(
                        f"Cannot convert '{arg}' to " +
                        f"{parameter.annotation.__name__} for parameter " +
                        f"'{parameter.name}' of command '{self.name}'")
                converted_args[name] = converted_arg
            else:
                converted_args[name] = arg
        return converted_args

    def _convert(self, value_str: str, type: type):
        if type is int or type is Optional[int]:
            return int(value_str)
        elif type is float or type is Optional[float]:
            return float(value_str)
        elif type is bool or type is Optional[bool]:
            if value_str == 'True' or value_str == 'true' or \
                    value_str == '1':
                return True
            elif value_str == 'False' or value_str == 'false' or \
                    value_str == '0':
                return False
            else:
                raise Exception("Invalid bool value")
        elif type is str or type is Optional[str] or type is None:
            return value_str
        else:
            raise Exception(f"Unknown type '{type}'")




import glob
from importlib.machinery import SourceFileLoader
from inspect import getmembers, isfunction
import json
import os
from types import ModuleType
from typing import Any, Callable, Optional

from cmdfns.command_cls import Command
from cmdfns.command_store import CommandStore


class FileSystemCommandStore(CommandStore):
    """Loads commands from Python files in the file system."""

    _search_path: str
    _index_path: Optional[str]
    _command_cache: dict[str, Command]
    _command_index: dict[str, str]

    def __init__(self, search_path: str, index_path: Optional[str] = None):
        """Create a CommandStore that looks for commands in the file system
        under the given search path.

        The store will search recursively for all Python files under
        search_path, then look for functions within them that have the
        @command decorator.

        If index_path is specified then it is used to write out an index
        file which is then loaded on subsequent runs to speed up command
        look-up.

        Parameters
        ----------
        search_path : str
            The path to search for Python files containing command functions.
        index_path : Optional[str], optional
            The path to write an index file to for speeding up subsequent
            command look-ups.
        """
        self._search_path = search_path
        self._index_path = index_path
        self._command_cache = {}
        self._command_index = {}

    def get_command(self, name: str) -> Optional[Command]:
        """Queries a command from the store.

        Parameters
        ----------
        name : str
            The name of the command to return.

        Returns
        -------
        Optional[Command]
            The command or None if not found.
        """
        command: Optional[Command] = self._find_command_from_cache(name)
        if not command:
            command = self._find_command_from_index(name)
        if not command:
            command = self._find_command_from_search_path(name)
        return command

    def get_commands(self) -> list[Command]:
        """Returns a list of all commands from the store.

        Returns
        -------
        list[Command]
            List of commands from the store.
        """
        self._populate_cache_from_search_path()
        return list(self._command_cache.values())

    def _find_command_from_cache(self, command_name: str) \
            -> Optional[Command]:
        return self._command_cache.get(command_name)

    def _find_command_from_index(self, command_name: str) \
            -> Optional[Command]:
        if not self._command_index:
            self._command_index = self._load_command_index()
        command_python_file = self._command_index.get(command_name)
        if not command_python_file or not os.path.exists(command_python_file):
            return None
        try:
            self._command_cache.update(
                self._load_commands_from_python_file(command_python_file))
        except Exception:
            pass
        return self._command_cache.get(command_name)

    def _find_command_from_search_path(self, command_name: str
                                       ) -> Optional[Command]:
        self._populate_cache_from_search_path()
        return self._command_cache.get(command_name)
    
    def _populate_cache_from_search_path(self) -> None:
        self._command_cache = self._find_commands_in_search_path()
        self._save_command_index()

    def _save_command_index(self) -> None:
        if self._index_path is None:
            return
        with open(self._index_path, 'w') as file:
            json.dump(
                {command.name: command.path
                    for command in self._command_cache.values()}, file)

    def _load_command_index(self) -> dict[str, str]:
        if self._index_path is None or not os.path.exists(self._index_path):
            return {}
        try:
            with open(self._index_path, 'r') as file:
                return json.load(file)
        except Exception:
            return {}

    def _find_commands_in_search_path(self) -> dict[str, Command]:
        commands: dict[str, Command] = {}
        python_files = self._find_python_files_in_search_path()
        for python_file in python_files:
            commands.update(self._load_commands_from_python_file(python_file))
        return commands

    def _load_commands_from_python_file(self, path: str) \
            -> dict[str, Command]:
        commands: dict[str, Command] = {}
        module = self._import_python_file(path)
        for (name, fn) in self._find_commands_in_module(module).items():
            commands[name] = Command(name, path, fn)
        return commands

    def _find_commands_in_module(self, module: ModuleType) \
            -> dict[str, Callable[..., Any]]:
        commands: dict[str, Callable[..., Any]] = {}
        for (_, member_obj) in getmembers(module):
            if isfunction(member_obj) and hasattr(member_obj, "_command"):
                command_name: str = member_obj._command["name"]  # type: ignore
                commands[command_name] = member_obj
        return commands

    def _import_python_file(self, path: str) -> ModuleType:
        return SourceFileLoader(
            self._python_file_path_to_module_name(path),
            path).load_module()

    def _python_file_path_to_module_name(self, path: str) -> str:
        filename: str = os.path.basename(path)
        if filename.endswith(".py"):
            filename = filename[:-3]
        return filename

    def _find_python_files_in_search_path(self) -> list[str]:
        python_files: list[str] = []
        for filename in glob.iglob(os.path.join(self._search_path, "**/*.py"),
                                   recursive=True):
            python_files.append(filename)
        return python_files




from typing import Any, Callable, Union


def command(fn_or_name: Union[str, Callable[..., Any]]) -> Callable[..., Any]:
    """Decorator for marking a function as a command that can be called from
    the command-line.

    Examples:

        @command
        def test_command():
            ...

        @command("a_custom_name")
        def another_test_command():
            ...

    Parameters
    ----------
    fn_or_name : Union[str, Callable[..., Any]]
        Optional custom command name, if not provided then the function
        name is used.

    Returns
    -------
    Callable[..., Any]
        The decorated function.
    """
    if isinstance(fn_or_name, Callable):
        fn_or_name._command = {"name": fn_or_name.__name__}  # type: ignore
        return fn_or_name
    else:
        def decorator_fn(fn: Callable[..., Any]) -> Callable[..., Any]:
            name: str = fn_or_name
            if name is None:
                name = fn.__name__
            fn._command = {"name": name}  # type: ignore
            return fn
        return decorator_fn




from typing import Optional

from cmdfns.command_cls import Command


class CommandStore:
    """Interface for querying commands from a store."""

    def get_command(self, name: str) -> Optional[Command]:
        """Queries a command from the store.

        Parameters
        ----------
        name : str
            The name of the command to return.

        Returns
        -------
        Optional[_Command]
            The command or None if not found.
        """
        raise Exception("Not implemented")

    def get_commands(self) -> list[Command]:
        """Returns a list of all commands from the store.

        Returns
        -------
        list[Command]
            List of commands from the store.
        """
        raise Exception("Not implemented")




import os
from typing import Any, Optional, Callable

from cmdfns.command_executor import CommandExecutor
from cmdfns.command_store import CommandStore
from cmdfns.dict_command_store import DictCommandStore
from cmdfns.file_system_command_store import FileSystemCommandStore


def main(search_path: Optional[str] = None,
         functions: Optional[dict[str, Callable[..., Any]]] = None,
         index_path: Optional[str] = None) -> Any:
    """Executes the command function specified by sys.argv, returning any
    result.

    The command function whose name matches sys.argv[1] is executed, with any
    subsequent values passed as positional arguments, except for those of the
    format "name=value" which are passed as keyword arguments. If no command
    function is found with the given name then usage help is printed to
    stdout.

    Exactly one of `search_path` or `functions` must be provided.

    If `search_path` is provided then this will search for functions decorated
    with `@cmdfns.command` in Python files in the same directory and
    recursively in any subdirectories of the given path.

    If `functions` is provided then it will be queried for command functions.

    If `index_path` is specified then it is used to write out an index
    file which is then loaded on subsequent runs to speed up command
    look-up.

    Parameters
    ----------
    search_path : Optional[str], optional
        The path to search for Python files containing command functions.
    index_path : Optional[str], optional
        The path to write an index file to for speeding up subsequent command
        look-ups.

    Returns
    -------
    Any
        The result of the command function, if any.
    """
    if search_path is not None and functions is not None or \
       search_path is None and functions is None:
        raise Exception("Exactly one of search_path or functions must be " +
                        "provided")
    command_store: CommandStore
    if search_path is not None:
        if os.path.isfile(search_path):
            search_path = os.path.dirname(search_path)
        search_path = os.path.abspath(search_path)
        command_store = FileSystemCommandStore(search_path, index_path)
    elif functions is not None:
        command_store = DictCommandStore(functions)
    else:
        raise Exception("Not implemented")
    return CommandExecutor(command_store).execute_command_from_argv()


def interactive_main(search_path: Optional[str] = None,
                     functions: Optional[dict[str, Callable[..., Any]]] = None,
                     index_path: Optional[str] = None,
                     exe_name: Optional[str] = None) -> None:
    """Interactively executes commands read from sys.stdin.

    Repeatedly reads commands from sys.stdin and executes them, with any
    values after each command name passed as positional arguments, except
    for those of the format "name=value" which are passed as keyword
    arguments. If no command function is found with the given name then usage
    help is printed to stdout. Terminates when the `quit` command is executed.

    Exactly one of `search_path` or `functions` must be provided.

    If `search_path` is provided then this will search for functions decorated
    with `@cmdfns.command` in Python files in the same directory and
    recursively in any subdirectories of the given path.

    If `functions` is provided then it will be queried for command functions.

    If `index_path` is specified then it is used to write out an index
    file which is then loaded on subsequent runs to speed up command
    look-up.

    Parameters
    ----------
    search_path : Optional[str], optional
        The path to search for Python files containing command functions.
    index_path : Optional[str], optional
        The path to write an index file to for speeding up subsequent command
        look-ups.
    exe_name : Optional[str], optional
        Optional executable name to use in help text (defaults to
        sys.argv[1]).
    """
    if search_path is not None and functions is not None or \
       search_path is None and functions is None:
        raise Exception("Exactly one of search_path or functions must be " +
                        "provided")
    command_store: CommandStore
    if search_path is not None:
        if os.path.isfile(search_path):
            search_path = os.path.dirname(search_path)
        search_path = os.path.abspath(search_path)
        command_store = FileSystemCommandStore(search_path, index_path)
    elif functions is not None:
        command_store = DictCommandStore(functions)
    else:
        raise Exception("Not implemented")
    return CommandExecutor(
        command_store, exe_name=exe_name).execute_command_from_stdin()


async def async_main(search_path: Optional[str] = None,
                     functions: Optional[dict[str, Callable[..., Any]]] = None,
                     index_path: Optional[str] = None,
                     exe_name: Optional[str] = None) -> Any:
    """Searches for command functions under the given search path and executes
    the one named in sys.argv.

    This function searches recursively for all Python files under search_path
    containing functions with the @cmdfns.command decorator. It then executes
    the function whose name matches sys.argv[1] and returns any result. If the
    command function is an async function then it is awaited.

    If no command function is found with the given name then usage help is
    printed to stdout.

    Subsequent values in sys.argv are passed as positional arguments, except
    for those of the format "name=value" which are passed as keyword
    arguments.

    Exactly one of `search_path` or `functions` must be provided.

    If `search_path` is provided then this will search for functions decorated
    with `@cmdfns.command` in Python files in the same directory and
    recursively in any subdirectories of the given path.

    If `functions` is provided then it will be queried for command functions.

    If `index_path` is specified then it is used to write out an index
    file which is then loaded on subsequent runs to speed up command
    look-up.

    Parameters
    ----------
    search_path : str
        The path to search for Python files containing command functions.
    index_path : Optional[str], optional
        The path to write an index file to for speeding up subsequent command
        look-ups.
    exe_name : Optional[str], optional
        Optional executable name to use in help text (defaults to
        sys.argv[1]).

    Returns
    -------
    Any
        The result of the command function, if any.
    """
    if search_path is not None and functions is not None or \
       search_path is None and functions is None:
        raise Exception("Exactly one of search_path or functions must be " +
                        "provided")
    command_store: CommandStore
    if search_path is not None:
        if os.path.isfile(search_path):
            search_path = os.path.dirname(search_path)
        search_path = os.path.abspath(search_path)
        command_store = FileSystemCommandStore(search_path, index_path)
    elif functions is not None:
        command_store = DictCommandStore(functions)
    else:
        raise Exception("Not implemented")
    return await CommandExecutor(
        command_store, exe_name=exe_name).execute_command_from_argv_async()


async def async_interactive_main(
        search_path: Optional[str] = None,
        functions: Optional[dict[str, Callable[..., Any]]] = None,
        index_path: Optional[str] = None) -> None:
    """Interactively executes commands read from sys.stdin.

    Repeatedly reads commands from sys.stdin and executes them, with any
    values after each command name passed as positional arguments, except
    for those of the format "name=value" which are passed as keyword
    arguments. If the command function is an async function then it is
    awaited. If no command function is found with the given name then usage
    help is printed to stdout. Terminates when the `quit` command is executed.

    Exactly one of `search_path` or `functions` must be provided.

    If `search_path` is provided then this will search for functions decorated
    with `@cmdfns.command` in Python files in the same directory and
    recursively in any subdirectories of the given path.

    If `functions` is provided then it will be queried for command functions.

    If `index_path` is specified then it is used to write out an index
    file which is then loaded on subsequent runs to speed up command
    look-up.

    Parameters
    ----------
    search_path : Optional[str], optional
        The path to search for Python files containing command functions.
    index_path : Optional[str], optional
        The path to write an index file to for speeding up subsequent command
        look-ups.
    """
    if search_path is not None and functions is not None or \
       search_path is None and functions is None:
        raise Exception("Exactly one of search_path or functions must be " +
                        "provided")
    command_store: CommandStore
    if search_path is not None:
        if os.path.isfile(search_path):
            search_path = os.path.dirname(search_path)
        search_path = os.path.abspath(search_path)
        command_store = FileSystemCommandStore(search_path, index_path)
    elif functions is not None:
        command_store = DictCommandStore(functions)
    else:
        raise Exception("Not implemented")
    return await CommandExecutor(
        command_store).execute_command_from_stdin_async()




from cmdfns.command import command
from cmdfns.main import main, interactive_main, async_main, \
    async_interactive_main

__all__ = ['command', 'main', 'interactive_main', 'async_main',
           'async_interactive_main']




from typing import Any, Callable, Optional

from cmdfns.command_cls import Command
from cmdfns.command_store import CommandStore


class DictCommandStore(CommandStore):
    """Returns commands from a dictionary."""

    _commands: dict[str, Command]

    def __init__(self, commands: dict[str, Callable[..., Any]]):
        """Create a CommandStore that returns command functions from the given
        dictionary.

        Parameters
        ----------
        commands : dict[str, Callable[..., Any]]
            Dictionary mapping command names to command functions.
        """
        self._commands = {}
        for name, function in commands.items():
            self._commands[name] = Command(name, name, function)

    def get_command(self, name: str) -> Optional[Command]:
        """Queries a command from the store.

        Parameters
        ----------
        name : str
            The name of the command to return.

        Returns
        -------
        Optional[_Command]
            The command or None if not found.
        """
        return self._commands.get(name)

    def get_commands(self) -> list[Command]:
        """Returns a list of all commands from the store.

        Returns
        -------
        list[Command]
            List of commands from the store.
        """
        return list(self._commands.values())




from setuptools import setup
import pathlib

here = pathlib.Path(__file__).parent.resolve()
long_description = (here / "README.md").read_text(encoding="utf-8")

setup(
  name="cmdfns",
  packages=["cmdfns"],
  package_data={"cmdfns": ["py.typed"]},
  version="0.8.0",
  license="MIT",
  description="Make Python functions callable from the command-line",
  long_description=long_description,
  long_description_content_type="text/markdown",
  author="Marc Hull",
  author_email="aozine@gmail.com",
  url="https://github.com/aozine/cmdfns",
  keywords=["command-line"],
  python_requires=">=3.7",
  install_requires=[
    "aioconsole==0.5.0"
  ],
  classifiers=[
    "Development Status :: 3 - Alpha",
    "Intended Audience :: Developers",
    "License :: OSI Approved :: MIT License",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: 3.10",
  ],
)




import aioconsole
import re
import shlex
import sys
import traceback
from typing import Any, Optional, Tuple

from cmdfns.command_cls import Command
from cmdfns.command_store import CommandStore
from cmdfns.invalid_argument import InvalidArgument


class CommandExecutor:
    """Handles execution of command functions."""

    _KEYWORD_ARG_REGEX = re.compile("([^ =]+)=(.+)")
    _store: CommandStore
    _exe_name: str

    def __init__(self, store: CommandStore, exe_name: Optional[str] = None):
        """Creates a CommandExecuter that looks for commands in the given
        store.

        Parameters
        ----------
        store : CommandStore
            The store to look for commands in.
        exe_name : Optional[str]
            Optional executable name to include in usage help, defaults to
            sys.argv[0].
        """
        self._store = store
        if exe_name:
            self._exe_name = exe_name
        else:
            self._exe_name = sys.argv[0]

    def execute_command_from_argv(self,
                                  argv: Optional[list[str]] = None) -> Any:
        """Execute a command from a list of command-line arguments.

        The first element of argv is assumed to be the Python script name and
        is ignored. The second element is treated as the command name.
        Subsequent elements are treated as command arguments.

        Parameters
        ----------
        argv : Optional[list[str]], optional
            Command-line argument list, if not specified then sys.argv is
            used.

        Returns
        -------
        Any
            The result of the command, if any.
        """
        if argv is None:
            argv = sys.argv
        if len(argv) < 2:
            self._print_usage(True)
            return
        self.execute_command_from_args(argv[1], argv[2:])

    def execute_command_from_args(self, command_name: str,
                                  args: list[str]) -> Any:
        """Execute a command given a command name and list of string
        arguments.

        Arguments are treated as positional arguments unless they are of the
        form "name=value", in which case they are treated as keyword
        arguments.

        If the command function is an async function then it is awaited.

        Parameters
        ----------
        command_name : str
            The name of the command to execute.
        args : list[str]
            A list of command arguments.

        Returns
        -------
        Any
            The result of the command, if any.
        """
        return self._execute_command_from_args(command_name, args, False)

    def execute_command_from_stdin(self):
        """Reads commands from stdin and executes them.

        Returns once the "quit" command has been executed.
        """
        self._print_usage(True)
        while True:
            argv: list[str] = shlex.split(input())
            command_name = argv[0]
            args = argv[1:]
            if command_name == "quit":
                return
            elif command_name == "help":
                self._print_help(args, True)
            else:
                try:
                    self.execute_command_from_args(command_name, args)
                except Exception:
                    traceback.print_exc()

    async def execute_command_from_argv_async(self,
                                              argv: Optional[list[str]] = None
                                              ) -> Any:
        """Execute a command from a list of command-line arguments.

        The first element of argv is assumed to be the Python script name and
        is ignored. The second element is treated as the command name.
        Subsequent elements are treated as command arguments.

        If the command function is an async function then it is awaited.

        Parameters
        ----------
        argv : Optional[list[str]], optional
            Command-line argument list, if not specified then sys.argv is
            used.

        Returns
        -------
        Any
            The result of the command, if any.
        """
        if argv is None:
            argv = sys.argv
        if len(argv) < 2:
            self._print_usage(False)
            return
        return await self.execute_command_from_args_async(argv[1], argv[2:])

    async def execute_command_from_args_async(self, command_name: str,
                                              args: list[str]) -> Any:
        """Execute a command given a command name and list of string
        arguments.

        Arguments are treated as positional arguments unless they are of the
        form "name=value", in which case they are treated as keyword
        arguments.

        If the command function is an async function then it is awaited.

        Parameters
        ----------
        command_name : str
            The name of the command to execute.
        args : list[str]
            A list of command arguments.

        Returns
        -------
        Any
            The result of the command, if any.
        """
        return await self._execute_command_from_args_async(
            command_name, args, False)

    async def execute_command_from_stdin_async(self):
        """Reads commands from stdin and executes them.

        If the command function is an async function then it is awaited.

        Returns once the "quit" command has been executed.
        """
        self._print_usage(True)
        while True:
            line: str = await aioconsole.ainput()
            argv: list[str] = shlex.split(line)
            command_name = argv[0]
            args = argv[1:]
            if command_name == "quit":
                return
            elif command_name == "help":
                self._print_help(args, True)
            else:
                try:
                    await self._execute_command_from_args_async(
                        command_name, args, True)
                except Exception:
                    traceback.print_exc()

    def _execute_command_from_args(self, command_name: str,
                                   args: list[str],
                                   interactive: bool) -> Any:
        if command_name == "help":
            self._print_help(args, interactive)
            return
        command: Optional[Command] = self._store.get_command(command_name)
        if not command:
            self._command_not_found(command_name, interactive)
            return
        positional_args, keyword_args = self._parse_args(args)
        try:
            return command.execute(positional_args, keyword_args)
        except InvalidArgument as e:
            print(str(e))

    async def _execute_command_from_args_async(self, command_name: str,
                                               args: list[str],
                                               interactive: bool) -> Any:
        if command_name == "help":
            self._print_help(args, interactive)
            return
        command: Optional[Command] = self._store.get_command(command_name)
        if not command:
            self._command_not_found(command_name, interactive)
            return
        positional_args, keyword_args = self._parse_args(args)
        try:
            return await command.execute_async(positional_args, keyword_args)
        except InvalidArgument as e:
            print(str(e))

    def _parse_args(self, args: list[str]) -> Tuple[list[str], dict[str, str]]:
        positional_args: list[str] = []
        keyword_args: dict[str, str] = {}
        for arg in args:
            match = self._KEYWORD_ARG_REGEX.match(arg)
            if match:
                keyword_args[match.group(1)] = match.group(2)
            else:
                positional_args.append(arg)
        return positional_args, keyword_args

    def _command_not_found(self, command_name: str, interactive: bool) -> None:
        print(f"Command '{command_name}' not found")
        print()
        self._print_usage(interactive)

    def _print_help(self, args: list[str], interactive: bool) -> None:
        if args:
            command_name = args[0]
            command: Optional[Command] = self._store.get_command(command_name)
            if not command:
                self._command_not_found(command_name, interactive)
                return
            self._print_command_help(command, interactive)
            return
        self._print_usage(interactive)

    def _print_command_help(self, command: Command, interactive: bool) -> None:
        print(f"Usage: {self._exe_name} {command.name} " +
              f"{command.args_string()}")
        if command.function.__doc__:
            print()
            print(command.function.__doc__)

    def _print_usage(self, interactive: bool) -> None:
        if interactive:
            print("Available commands:")
        else:
            print(f"Usage: {self._exe_name} COMMAND [ARGS]")
            print()
            print("Where COMMAND is one of:")
        for command in self._store.get_commands():
            print(f"  {command.name} {command.args_string()}")
        print()
        if interactive:
            print("Use 'quit' to quit or 'help COMMAND' for " +
                  "command-specific help")
        else:
            print(f"Use '{self._exe_name} help COMMAND' for " +
                  "command-specific help")
